==> BestCrypt-2.0.7.tar.gz: BestCrypt-2.0.7/certificate/CertificateLib/x509pkts/base64funcs.cpp <==
#include "base64funcs.h"
#include <string.h>
#include <assert.h>

//
//	ASCII Armor
//	6.2.     Forming ASCII Armor                                      42
//	RFC 2440                 OpenPGP Message Format            November 1998
// 
// 
//          +--first octet--+-second octet--+--third octet--+
//          |7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|
//          +-----------+---+-------+-------+---+-----------+
//          |5 4 3 2 1 0|5 4 3 2 1 0|5 4 3 2 1 0|5 4 3 2 1 0|
//          +--1.index--+--2.index--+--3.index--+--4.index--+
// 
//    Each 6-bit group is used as an index into an array of 64 printable
//    characters from the table below. The character referenced by the
//    index is placed in the output string.
// 
//	   Special processing is performed if fewer than 24 bits are available
//	   at the end of the data being encoded. There are three possibilities:
//	
//	    1. The last data group has 24 bits (3 octets). No special
//	       processing is needed.
//	
//	    2. The last data group has 16 bits (2 octets). The first two 6-bit
//	       groups are processed as above. The third (incomplete) data group
//	       has two zero-value bits added to it, and is processed as above.
//	       A pad character (=) is added to the output.
//	
//	    3. The last data group has 8 bits (1 octet). The first 6-bit group
//	       is processed as above. The second (incomplete) data group has
//	       four zero-value bits added to it, and is processed as above. Two
//	       pad characters (=) are added to the output.
//	

const char *base64T = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
const int base64TSize = 64;

static bool findChar( const char *T, int Q, char ch )
{
	for( int i = Q; i--; )
	{
		if ( T[i] == ch )
			return true;
	}

	return false;
}

//static bool isValidBase64Symbol( char ch )
//{
//	return findChar( base64T, base64TSize, ch );
//}

const char invisibleT[] = { '\n', '\r', ' ', '\t' };
						// 0x0A, 0x0D, 0x20, 0x09 };

#ifndef maxLen
#define maxLen( _XX ) ( sizeof( _XX ) / sizeof( *_XX ) )
#endif

static bool isInvisible( char ch ) { return findChar( invisibleT, maxLen( invisibleT ), ch ); }

static int findString( const BYTE *pkt, int pktSize, const char *str )
{
	int checkLength = strlen( str );
	int checkCount = 0;
	int pos = 0;

	for(; pos + checkCount < pktSize ;)
	{
		if ( pkt[ pos + checkCount ] == str[ checkCount ] )
		{
			if ( ++checkCount == checkLength )
			{
				return pos;
			}
		}
		else	// Reset counter
		{
			checkCount = 0;
			pos ++;
		}
	}

	return -1;
}

//
//	It just skips header lines.
//	The header is double new line terminated.
//	The problem are.
//	Originally the double new line is 0a 0a. But email clients modify text.
//	It may be 0d 0a 0d 0a or even 0d 0a 20 0d 0a. They add any invisible characters.
//	The function skips the invisible characters when the double new line is finding.
//
static int getBodyPos( const BYTE *bf, int size )
{
	bool bFirstFound = false;

	for( int i = 0 ; i < size ; i++ )
	{
		if ( bf[i] == 0x0A )
		{
			if ( bFirstFound )
				return i + 1;
			else
				bFirstFound = true;
		}
		else
//	The header string allows all readable symbols.
//		if ( isValidBase64Symbol( bf[i] ) )
//			bFirstFound = false;
//		else
		if ( isInvisible( bf[i] ) )
			continue;
		else 
		{
			bFirstFound = false;
//			assert( false );
		}
	}

	return 0;
}

int findEncodedBlock(
		const BYTE **pBlock,		// OUT - Pointer to byte position of encoded block
		int *pBlockSize,	// OUT - Pointer to finded size of encoded block
		const BYTE *pkt,	// IN - the packet with encoded block to be find.
		int pktSize,		// IN - the packet size
		const char *szBlockStart,
		const char *szBlockEnd,
		const char **pArmorHeaders// OUT pointer to position of Armor headers
							// the headers are separated by single 0x0A symbol
							// and are terminated by double 0x0A symbols ( a blank line )
		)
{
	int posStart = findString( pkt, pktSize, szBlockStart );
	
	if ( posStart == -1 )
		return ppe_invalidBlock;

	posStart += strlen( szBlockStart );

	int posEnd = findString( pkt + posStart, pktSize - posStart, szBlockEnd );

	if ( posEnd == -1 )
		return ppe_invalidBlock;

	////	
	////	Finds header lines in armored message.
	////
	//BYTE blankLine[3] = { 0x0A, 0x0A, 0 };

	//int posBody = findString( pkt + posStart, posEnd, (char *)blankLine );

	//if ( posBody == -1 )
	//{
	//	BYTE blankLine2[5] = { 0x0D, 0x0A, 0x0D, 0x0A, 0 };
	//	posBody = findString( pkt + posStart, posEnd, (char *)blankLine2 );
	//}

	//if ( posBody == -1 ) // No headers found
	//	posBody = posStart;
	//else
	//{
	//	posBody += strlen( (char *)blankLine );
	//	posEnd -= posBody; // Position from body position
	//	posBody = posStart + posBody; // Posotion from pkt begin
	//	
	//}

	int posBody = getBodyPos( pkt + posStart, posEnd );

	if ( posBody <= 0 ) // No headers found
		posBody = posStart;
	else
	{
//		posBody += strlen( (char *)blankLine );
		posEnd -= posBody; // Position from body position
		posBody = posStart + posBody; // Posotion from pkt begin
		
	}

	if ( pArmorHeaders != NULL )
		*pArmorHeaders = (char *)pkt + posStart + 1; // +1 skips 0x0A symbols before headers.

	*pBlock = pkt + posBody;
	*pBlockSize = posEnd;

	return ppe_ok;
}

static int getCode( BYTE ch )
{
	int code = base64TSize + 1;

	for (; code--;)
	{
		if( base64T[ code ] == ch )
			break;
	}

	return code;
}

static int getDecodedSizeAndCheck( 
		const BYTE *encBlock,		// IN - encoded buffer
		int encBlockSize )
{
	int code;
	
	int size = 0;
	int padQ = 0;
	int count = 0;

	bool newLine = false;

	for(; count < encBlockSize; count++ )
	{
		BYTE ch = encBlock[ count ];

		code = getCode( ch );

		if (( ch == 0x0A ) || ( ch == 0x0D ) )
			newLine = true;

		if ( code == base64TSize )
		{
			if ( newLine )	break; // We encountered with check summ.

			padQ++;
		}
		else
		if ( code >= 0 )
		{
			newLine = false;
			size++;
		}
	}

	size = ( size / 4 ) * 3;

	if ( padQ == 1 )
		size += 2;
	else
	if ( padQ == 2 )
		size += 1;
	else
	if ( padQ != 0 )
		return (-1);

	return size;
}

int base64Decode(
		BYTE *decBlock,		// OUT - buffer for decoded data.
		int *pDecBlockSize,	// IN - pointer for decoded size, if size of 
							// the output block is too small return ppe_bufferTooSmall
							// OUT - required or really filled size 
		const BYTE *encBlock,		// IN - encoded buffer
		int encBlockSize )
{
	int size = getDecodedSizeAndCheck( encBlock, encBlockSize );

	if ( size == -1 )
		return ppe_invalidBlock;

	if ( size > *pDecBlockSize )
	{
		*pDecBlockSize = size;
		return ppe_bufferTooSmall;
	}

	int posD = 0, posE = 0;

	UINT dw24 = 0;

	int dwPos = 0;

	bool newLine = false;

	for(; posE < encBlockSize; )
	{
		BYTE ch = encBlock[ posE++ ];

		if ( ( ch == 0x0A ) || ( ch == 0x0D ) )
		{
			newLine = true;
			continue;
		}

		int code = getCode( ch );

		if ( code < 0 )
			continue;

		if ( code == base64TSize ) // pad -- '='
		{
			if ( newLine )	// 0x0A, "=QWER" It is a check summ just after newline
				break;

			dw24 = ( dw24 << 6 );
			continue;
		}
		else
		{
			dw24 = ( dw24 << 6 ) | (UINT) code;
			newLine = false;
		}

		if ( ++dwPos == 4 )
		{
			assert( posD + 2 < *pDecBlockSize );

			dwPos = 0;
			if ( posD + 2 >=  size )
				return ppe_invalidBlock;

			decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
			decBlock[ posD++ ] = (BYTE)( dw24 >> 8  );
			decBlock[ posD++ ] = (BYTE)( dw24       );

			dw24 = 0;
		}
	}


	if ( dwPos == 3 )
	{
		assert( posD + 1 < *pDecBlockSize );

		//
		//	03.06.2013
		//	If an invalid symbol was encounterred asertion above works, so check below is added
		//
		if ( posD + 1 >= size )
			return ppe_invalidBlock;

		decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
		decBlock[ posD++ ] = (BYTE)( dw24 >> 8  );
	}
	else
	if ( dwPos == 2 )
	{
		assert( posD < *pDecBlockSize );

		if ( posD >= size )
			return ppe_invalidBlock;

		decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
	}

	*pDecBlockSize = posD;

	return ppe_ok;
}

int base64Encode(
		BYTE *encBlock,		// OUT - buffer for encoded data.
		int *pEncBlockSize,	// IN - pointer for encoded size, if size of 
							// the output block is too small return ppe_bufferTooSmall
							// OUT - required or really filled size 
		const BYTE *decBlock,		// IN - plain text buffer
		int decBlockSize,

		const char *szBlockStart,	// If the strings are not NULL,
		const char *szBlockEnd,		// so encloses the encoded block by the strings

		const char *szVerInfo		// PGP allows version information string before real armored data
	)
{
	int encodedSize = ( decBlockSize / 3 ) * 4;
	int tail = decBlockSize % 3;

	//int padded = 0;

	//if ( tail == 2 )
	//	padded = 1;
	//else
	//if ( tail == 1 )
	//	padded = 2;

	if ( tail )
		encodedSize += 4;

	int lines = ( encodedSize + 63  ) / 64;

	bool useBoundStrings = (( szBlockStart != NULL ) && ( szBlockEnd != NULL ));

	if ( useBoundStrings )
	{
		lines += 2;
		encodedSize += ( strlen( szBlockStart ) + strlen( szBlockEnd ) );
	}

	bool useVersionInfo = ( ( szVerInfo != NULL ) && ( strlen(szVerInfo) != 0 ) );
	const char *szVersion = "Version: ";

	if ( useVersionInfo )
	{
		lines += 2;
		encodedSize += ( strlen( szVerInfo ) + strlen( szVersion ) );// +  2 * strlen( "\r\n" ) );
	}

	encodedSize += lines * 2; // "\r\n"

	if ( *pEncBlockSize < encodedSize )
	{
		*pEncBlockSize = encodedSize;
		return ppe_bufferTooSmall;
	}

	int bl; // block number

	int ePos = 0;

	if ( useBoundStrings )
	{
		int len = strlen( szBlockStart );
		memcpy( encBlock + ePos, szBlockStart, len );
		ePos += len;
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}
	
	if ( useVersionInfo )
	{
		int len = strlen( szVersion );
		memcpy( encBlock + ePos, szVersion, len );
		ePos += len;

		len = strlen( szVerInfo );
		memcpy( encBlock + ePos, szVerInfo, len );
		ePos += len;

		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}

	for ( bl = 0; bl < ( decBlockSize / 3 );  )
	{
		UINT dw24 =  UINT( decBlock[ 3 * bl + 2 ] )         | 
					( UINT( decBlock[ 3 * bl + 1 ] ) << 8  ) |
					( UINT( decBlock[ 3 * bl     ] ) << 16 );

		encBlock[ ePos++ ] = base64T[ ( dw24 >> 18 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >> 12 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >>  6 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24       ) & 0x3F ];

		bl++;

		if ( ( bl % 16 ) == 0 ) // new line required;
		{
			encBlock[ ePos++ ] = '\r';
			encBlock[ ePos++ ] = '\n';
		}
	}

	if ( tail )
	{
		UINT dw24 =  UINT( decBlock[ 3 * bl ] ) << 16;
		if ( tail == 2 )
			dw24 |= ( UINT( decBlock[ 3 * bl + 1 ] ) << 8 );

		encBlock[ ePos++ ] = base64T[ ( dw24 >> 18 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >> 12 ) & 0x3F ];

		if ( tail == 2 )
			encBlock[ ePos++ ] = base64T[ ( dw24 >> 6 ) & 0x3F ];
		else
			encBlock[ ePos++ ] = base64T[ 0x40 ];

		encBlock[ ePos++ ] = base64T[ 0x40 ];

		bl++;

		if ( ( bl % 16 ) == 0 ) // new line required;
		{
			encBlock[ ePos++ ] = '\r';
			encBlock[ ePos++ ] = '\n';
		}
	}

	if ( ( bl % 16 ) != 0 )
	{
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}

	if ( useBoundStrings )
	{
		int len = strlen( szBlockEnd );
		memcpy( encBlock + ePos, szBlockEnd, len );
		ePos += len;

		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}

	if ( ePos != encodedSize )
		return ppe_internalError;

	*pEncBlockSize = encodedSize;

	return ppe_ok;
}

==> BestCrypt-2.0.14.tar.gz: BestCrypt-2.0.14/certificate/CertificateLib/x509pkts/base64funcs.cpp <==
#include "base64funcs.h"
#include <string.h>

//
//	ASCII Armor
//	6.2.     Forming ASCII Armor                                      42
//	RFC 2440                 OpenPGP Message Format            November 1998
// 
// 
//          +--first octet--+-second octet--+--third octet--+
//          |7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|
//          +-----------+---+-------+-------+---+-----------+
//          |5 4 3 2 1 0|5 4 3 2 1 0|5 4 3 2 1 0|5 4 3 2 1 0|
//          +--1.index--+--2.index--+--3.index--+--4.index--+
// 
//    Each 6-bit group is used as an index into an array of 64 printable
//    characters from the table below. The character referenced by the
//    index is placed in the output string.
// 
//	   Special processing is performed if fewer than 24 bits are available
//	   at the end of the data being encoded. There are three possibilities:
//	
//	    1. The last data group has 24 bits (3 octets). No special
//	       processing is needed.
//	
//	    2. The last data group has 16 bits (2 octets). The first two 6-bit
//	       groups are processed as above. The third (incomplete) data group
//	       has two zero-value bits added to it, and is processed as above.
//	       A pad character (=) is added to the output.
//	
//	    3. The last data group has 8 bits (1 octet). The first 6-bit group
//	       is processed as above. The second (incomplete) data group has
//	       four zero-value bits added to it, and is processed as above. Two
//	       pad characters (=) are added to the output.
//	

const char *base64T = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
const int base64TSize = 64;

static bool findChar( const char *T, int Q, char ch )
{
	for( int i = Q; i--; )
	{
		if ( T[i] == ch )
			return true;
	}

	return false;
}

//static bool isValidBase64Symbol( char ch )
//{
//	return findChar( base64T, base64TSize, ch );
//}

const char invisibleT[] = { '\n', '\r', ' ', '\t' };
						// 0x0A, 0x0D, 0x20, 0x09 };

#ifndef maxLen
#define maxLen( _XX ) ( sizeof( _XX ) / sizeof( *_XX ) )
#endif

static bool isInvisible( char ch ) { return findChar( invisibleT, maxLen( invisibleT ), ch ); }

static int findString( const BYTE *pkt, int pktSize, const char *str )
{
	int checkLength = strlen( str );
	int checkCount = 0;
	int pos = 0;

	for(; pos + checkCount < pktSize ;)
	{
		if ( pkt[ pos + checkCount ] == str[ checkCount ] )
		{
			if ( ++checkCount == checkLength )
			{
				return pos;
			}
		}
		else	// Reset counter
		{
			checkCount = 0;
			pos ++;
		}
	}

	return -1;
}

//
//	It just skips header lines.
//	The header is double new line terminated.
//	The problem are.
//	Originally the double new line is 0a 0a. But email clients modify text.
//	It may be 0d 0a 0d 0a or even 0d 0a 20 0d 0a. They add any invisible characters.
//	The function skips the invisible characters when the double new line is finding.
//

// 22.06.2016
// Email adds double newlines like below, if we found it, so it definitely is body begin.
static const char szDoubleNL[] = { 0x0d, 0x0a, 0x0d, 0x0a, 0x20, 0x0d, 0x0a, 0x0d, 0x0a, 0x00 };

static int getBodyPos( const BYTE *bf, int size )
{
	bool bFirstFound = false;

	const char *szBody = strstr( (const char*)bf, szDoubleNL );
	if ( szBody != NULL )
	{
		int pos = (int)(szBody - (const char*)bf);
		pos += strlen( szDoubleNL );

		if ( pos < size )
			return pos;
	}

	for( int i = 0 ; i < size ; i++ )
	{
		if ( bf[i] == 0x0A )
		{
			if ( bFirstFound )
				return ++i;
			else
				bFirstFound = true;
		}
		else
//	The header string allows all readable symbols.
//		if ( isValidBase64Symbol( bf[i] ) )
//			bFirstFound = false;
//		else
		if ( isInvisible( bf[i] ) )
			continue;
		else 
		{
			bFirstFound = false;
		}
	}

	return 0;
}

int findEncodedBlock(
		const BYTE **pBlock,		// OUT - Pointer to byte position of encoded block
		int *pBlockSize,	// OUT - Pointer to finded size of encoded block
		const BYTE *pkt,	// IN - the packet with encoded block to be find.
		int pktSize,		// IN - the packet size
		const char *szBlockStart,
		const char *szBlockEnd,
		const char **pArmorHeaders// OUT pointer to position of Armor headers
							// the headers are separated by single 0x0A symbol
							// and are terminated by double 0x0A symbols ( a blank line )
		)
{
	int posStart = findString( pkt, pktSize, szBlockStart );
	
	if ( posStart == -1 )
		return ppe_invalidBlock;

	posStart += strlen( szBlockStart );

	int posEnd = findString( pkt + posStart, pktSize - posStart, szBlockEnd );

	if ( posEnd == -1 )
		return ppe_invalidBlock;

	////	
	////	Finds header lines in armored message.
	////
	//BYTE blankLine[3] = { 0x0A, 0x0A, 0 };

	//int posBody = findString( pkt + posStart, posEnd, (char *)blankLine );

	//if ( posBody == -1 )
	//{
	//	BYTE blankLine2[5] = { 0x0D, 0x0A, 0x0D, 0x0A, 0 };
	//	posBody = findString( pkt + posStart, posEnd, (char *)blankLine2 );
	//}

	//if ( posBody == -1 ) // No headers found
	//	posBody = posStart;
	//else
	//{
	//	posBody += strlen( (char *)blankLine );
	//	posEnd -= posBody; // Position from body position
	//	posBody = posStart + posBody; // Posotion from pkt begin
	//	
	//}

	int posBody = getBodyPos( pkt + posStart, posEnd );

	if ( posBody <= 0 ) // No headers found
		posBody = posStart;
	else
	{
//		posBody += strlen( (char *)blankLine );
		posEnd -= posBody; // Position from body position
		posBody = posStart + posBody; // Posotion from pkt begin
	}

	if ( pArmorHeaders != NULL )
	{
		*pArmorHeaders = (char *)pkt + posStart + 1; // +1 skips 0x0A symbols before headers.
		while( isInvisible( *pArmorHeaders[0] ) )
			pArmorHeaders++;
	}

	*pBlock = pkt + posBody;
	*pBlockSize = posEnd;

	return ppe_ok;
}

static int getCode( BYTE ch )
{
	int code = base64TSize + 1;

	for (; code--;)
	{
		if( base64T[ code ] == ch )
			break;
	}

	return code;
}

static int getDecodedSizeAndCheck( 
		const BYTE *encBlock,		// IN - encoded buffer
		int encBlockSize )
{
	int code;
	
	int size = 0;
	int padQ = 0;
	int count = 0;

	bool newLine = false;

	for(; count < encBlockSize; count++ )
	{
		BYTE ch = encBlock[ count ];

		code = getCode( ch );

		if (( ch == 0x0A ) || ( ch == 0x0D ) )
			newLine = true;

		if ( code == base64TSize )
		{
			if ( newLine )
				break; // We encountered with check summ.

			padQ++;
		}
		else
		if ( code >= 0 )
		{
			newLine = false;
			size++;
		}
	}

	size = ( size / 4 ) * 3;

	if ( padQ == 1 )
		size += 2;
	else
	if ( padQ == 2 )
		size += 1;
	else
	if ( padQ != 0 )
		return (-1);

	return size;
}

static bool isEmptySymbol( char ch )
{
	char T[] = { '\r', '\n', '\t', ' ' };

	for( int i = sizeof( T ); i--; )
	{
		if ( ch == T[i] )
			return true;
	}

	return false;
}

// RFC 4880                 OpenPGP Message Format            November 2007
// 6.1.  An Implementation of the CRC-24 in "C"

#define CRC24_INIT 0xB704CEL
#define CRC24_POLY 0x1864CFBL

typedef long crc24;
crc24 crc_octets( const unsigned char *octets, size_t len)
{
	  crc24 crc = CRC24_INIT;
	  int i;
	  while (len--) {
		  crc ^= (*octets++) << 16;
		  for (i = 0; i < 8; i++) {
			  crc <<= 1;
			  if (crc & 0x1000000)
				  crc ^= CRC24_POLY;
		  }
	  }
	  return crc & 0xFFFFFFL;
}


/*
-----BEGIN PGP MESSAGE-----
Version: GnuPG v1

iQEcBAABAgAGBQJX8jwCAAoJECeTI+lqIvfDqQYH/j6NWXqJFPdbxKDFYtGohuvx
s/7UQ18Hg3pU4MFpSfxinQVH/SF7FBlG/wpQxFvOgiH39Nu4nyi5XR7PYwbKoq2/
HFnjtxm6XQ7zkvAPlmZh4s+cE64e4BsYjXDXl2AJuxHYtHcZLSk4iJidTV7b6X76
RorJC6CaGJnWM+soP+kzUyuOcV7nXdf1RJhhBzkZ+zCmJWGrtbutJhF6gcQClU8s
TusETgMx8gEIoaVzzOCz35/4yOEePMPh7d5AXLnKSB0OOIIPaxsvULSbPM0El/7T
uaQI5TPT081z9S9q7noTlre7z8npZuDL9TnXpo/lCc0bZ33pYDTMtnz5eaB5fUg=
=hEDy
-----END PGP MESSAGE-----
*/


static bool readCRC( crc24 &dw24, const BYTE *encBlock, int encBlockSize )
{
	if ( encBlockSize < 4 ) 
		return false;

	// big endian
	dw24 = 0;
	for( int i = 0; i < 4; i++ )
	{
		BYTE ch = encBlock[ i ];
		UINT code = getCode( ch );
		dw24 = ( dw24 << 6 ) | code;
	}

	return true;
}

int base64Decode(
		BYTE *decBlock,		// OUT - buffer for decoded data.
		int *pDecBlockSize,	// IN - pointer for decoded size, if size of 
							// the output block is too small return ppe_bufferTooSmall
							// OUT - required or really filled size 
		const BYTE *encBlock,		// IN - encoded buffer
		int encBlockSize )
{
	int size = getDecodedSizeAndCheck( encBlock, encBlockSize );

	if ( size == -1 )
		return ppe_invalidBlock;

	if ( size > *pDecBlockSize )
	{
		*pDecBlockSize = size;
		return ppe_bufferTooSmall;
	}

	int posD = 0, posE = 0;

	UINT dw24 = 0;

	int dwPos = 0;

	bool newLine = false;

	crc24 crcRead = 0;
	bool bHasCRC = false;

	for(; posE < encBlockSize; )
	{
		BYTE ch = encBlock[ posE++ ];

		if ( ( ch == 0x0A ) || ( ch == 0x0D ) )
		{
			newLine = true;
			while( isEmptySymbol( encBlock[posE] ) )
				posE++;

			continue;
		}

		int code = getCode( ch );

		if ( code < 0 )
			continue;

		if ( code == base64TSize ) // pad -- '='
		{
			if ( newLine )	// 0x0A, "=QWER" It is a check summ just after newline
			{
				if ( !readCRC( crcRead, encBlock + posE, encBlockSize - posE ) )
					return ppe_inputDataIsCorrupted;

				bHasCRC = true;
				break;
			}

			dw24 = ( dw24 << 6 );
			continue;
		}
		else
		{
			dw24 = ( dw24 << 6 ) | (UINT) code;
			newLine = false;
		}

		if ( ++dwPos == 4 )
		{
			if ( posD + 2 >= *pDecBlockSize ) // v.1.04.8.3 - 04.10.2016
				return ppe_invalidBlock;

			dwPos = 0;
			if ( posD + 2 >=  size )
				return ppe_invalidBlock;

			decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
			decBlock[ posD++ ] = (BYTE)( dw24 >> 8  );
			decBlock[ posD++ ] = (BYTE)( dw24       );

			dw24 = 0;
		}
	}


	if ( dwPos == 3 )
	{
		if ( posD + 1 >= *pDecBlockSize ) // v.1.04.8.3 - 04.10.2016
			return ppe_invalidBlock;

		//
		//	03.06.2013
		//	If an invalid symbol was encounterred asertion above works, so check below is added
		//
		if ( posD + 1 >= size )
			return ppe_invalidBlock;

		decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
		decBlock[ posD++ ] = (BYTE)( dw24 >> 8  );
	}
	else
	if ( dwPos == 2 )
	{
		if ( posD >= *pDecBlockSize )	// v.1.04.8.3 - 04.10.2016
			return ppe_invalidBlock;

		if ( posD >= size )
			return ppe_invalidBlock;

		decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
	}

	*pDecBlockSize = posD;

	if ( bHasCRC  )
	{
		crc24 crcCount = crc_octets( decBlock, posD );

		if ( crcRead != crcCount )
			return ppe_inputDataIsCorrupted;
	}

	return ppe_ok;
}

int base64Encode(
		BYTE *encBlock,		// OUT - buffer for encoded data.
		int *pEncBlockSize,	// IN - pointer for encoded size, if size of 
							// the output block is too small return ppe_bufferTooSmall
							// OUT - required or really filled size 
		const BYTE *decBlock,		// IN - plain text buffer
		int decBlockSize,

		const char *szBlockStart,	// If the strings are not NULL,
		const char *szBlockEnd,		// so encloses the encoded block by the strings

		const char *szVerInfo		// PGP allows version information string before real armored data
	)
{
	int encodedSize = ( decBlockSize / 3 ) * 4;
	int tail = decBlockSize % 3;

	//int padded = 0;

	//if ( tail == 2 )
	//	padded = 1;
	//else
	//if ( tail == 1 )
	//	padded = 2;

	if ( tail )
		encodedSize += 4;

	int lines = ( encodedSize + 63  ) / 64;

	bool bUseCheckSumm = true;

	if ( bUseCheckSumm )
	{
		// '=QWER'
		// '\r\n'
		encodedSize += 5; 
		lines += 1;
	}

	bool useBoundStrings = (( szBlockStart != NULL ) && ( szBlockEnd != NULL ));

	if ( useBoundStrings )
	{
		lines += 2;
		encodedSize += ( strlen( szBlockStart ) + strlen( szBlockEnd ) );
	}

	bool useVersionInfo = ( ( szVerInfo != NULL ) && ( strlen(szVerInfo) != 0 ) );
	const char *szVersion = "Version: ";

	if ( useVersionInfo )
	{
		lines += 2;
		encodedSize += ( strlen( szVerInfo ) + strlen( szVersion ) );// +  2 * strlen( "\r\n" ) );
	}

	encodedSize += lines * 2; // "\r\n"

	if ( *pEncBlockSize < encodedSize )
	{
		*pEncBlockSize = encodedSize;
		return ppe_bufferTooSmall;
	}

	int bl; // block number

	int ePos = 0;

	if ( useBoundStrings )
	{
		int len = strlen( szBlockStart );
		memcpy( encBlock + ePos, szBlockStart, len );
		ePos += len;
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}
	
	if ( useVersionInfo )
	{
		int len = strlen( szVersion );
		memcpy( encBlock + ePos, szVersion, len );
		ePos += len;

		len = strlen( szVerInfo );
		memcpy( encBlock + ePos, szVerInfo, len );
		ePos += len;

		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}

	for ( bl = 0; bl < ( decBlockSize / 3 );  )
	{
		UINT dw24 =  UINT( decBlock[ 3 * bl + 2 ] )         | 
					( UINT( decBlock[ 3 * bl + 1 ] ) << 8  ) |
					( UINT( decBlock[ 3 * bl     ] ) << 16 );

		encBlock[ ePos++ ] = base64T[ ( dw24 >> 18 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >> 12 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >>  6 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24       ) & 0x3F ];

		bl++;

		if ( ( bl % 16 ) == 0 ) // new line required;
		{
			encBlock[ ePos++ ] = '\r';
			encBlock[ ePos++ ] = '\n';
		}
	}

	if ( tail )
	{
		UINT dw24 =  UINT( decBlock[ 3 * bl ] ) << 16;
		if ( tail == 2 )
			dw24 |= ( UINT( decBlock[ 3 * bl + 1 ] ) << 8 );

		encBlock[ ePos++ ] = base64T[ ( dw24 >> 18 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >> 12 ) & 0x3F ];

		if ( tail == 2 )
			encBlock[ ePos++ ] = base64T[ ( dw24 >> 6 ) & 0x3F ];
		else
			encBlock[ ePos++ ] = base64T[ 0x40 ];

		encBlock[ ePos++ ] = base64T[ 0x40 ];

		bl++;

		if ( ( bl % 16 ) == 0 ) // new line required;
		{
			encBlock[ ePos++ ] = '\r';
			encBlock[ ePos++ ] = '\n';
		}
	}
	
	if ( ( bl % 16 ) != 0 )
	{
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}

	if ( bUseCheckSumm )
	{
		crc24 chk24 = crc_octets( decBlock, decBlockSize );

		encBlock[ ePos++ ] = '=';

		for( int i = 0; i<4; i++ )
		{
			int shift = ( ( 3 - i) * 6 );
			UINT code = ( chk24 >> shift )& 0x3F;
			encBlock[ePos++] = base64T[ code ];
		}

		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';

		crc24 crcRead;
		if ( !readCRC( crcRead, encBlock + ePos - 6, 6 ) )
			return ppe_internalError;

		if ( crcRead != chk24 )
			return ppe_internalError;
	}

	if ( useBoundStrings )
	{
		int len = strlen( szBlockEnd );
		memcpy( encBlock + ePos, szBlockEnd, len );
		ePos += len;

		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}

	if ( ePos != encodedSize )
		return ppe_internalError;

	*pEncBlockSize = encodedSize;

	return ppe_ok;
}
